import json
import logging
import threading

import spread_core.mqtt as mqtt
from spread_core.bam import generator as generator
from spread_core.bam.entities import Entity
from spread_core.bam.multi_command import MultiCommand
from spread_core.errors.project_errors import ProjectError, JocketError, InvalidDevice, InitError, ClassifierError, \
    CommandError
from spread_core.protocols.dali.exceptions import CommandGenerateError
from spread_core.tools.manager_interface import ManagerOfBroker

CONTROL = 4
COMMAND = 3
TUNER = 2
INFO = 1
SURVEY = 0


class Manager(Entity, ManagerOfBroker):
    STATUS_STOPPED = 'Stopped'
    STATUS_RUNNING = 'Running'
    STATUS_INFO = 'Info'
    STATUS_SURVEY = 'Survey'

    _has_poll_rate = True
    _server_id = None
    _key = None
    _bus_id = None
    _pollRate = None
    _survey_timer = None
    _survey_event = threading.Event()
    _step_event = threading.Event()
    _event = threading.Event()
    _external_cmd = threading.Event()
    INTERRUPT_SURVEY = True

    server_id = property(fget=lambda self: self._server_id)

    def __init__(self, p_id,  m_id):
        Entity.__init__(self, m_id)
        self._stopped = False
        self._send_interface = None
        self._providers = dict()
        self._addresses = dict()
        self._subgineries = dict()
        self._engineries = dict()
        self._project_id = p_id
        self.actions = {CONTROL: [], COMMAND: [], TUNER: [], INFO: [], SURVEY: []}
        self.ready_timer = None
        self.bus = None
        self.name = ''

    def _sorted_info(self):
        return self._providers

    @property
    def send_interface(self):
        return self._send_interface

    @send_interface.setter
    def send_interface(self, interface):
        if interface:
            interface.set_event_handler(self._event)
            interface.set_external_cmd_handler(self._external_cmd)
        self._send_interface = interface

    @property
    def project_id(self):
        return self._project_id

    @property
    def providers(self):
        """returns copy of providers"""
        return self._providers.copy()

    def get_provider(self, provider_id):
        """returns provider by id"""
        return self._providers.get(provider_id, None)

    def providers_by_address(self, short_address) -> list:
        """returns provider by address"""
        provider_ids = self._addresses.get(short_address, [])
        return [self._providers.get(provider_id, None) for provider_id in provider_ids]

    @property
    def addresses(self):
        """returns copy of addresses"""
        return self._addresses.copy()

    def publish(self, topic, data, retain=False, **kwargs):
        ManagerOfBroker.publish(self, topic, data, retain=retain, **kwargs)

    def set_addresses(self, address: int, participants: list):
        """set participants of address"""
        if isinstance(participants, (list, tuple)):
            self._addresses[address] = list(participants)

    def parse(self, data):
        if data['type'] != self.__class__.__name__:
            raise InitError(f'Manager type required as "{self.__class__.__name__}" but set "{data["type"]}"')
        self._server_id = data['serverID']
        self.bus = data['attributes'].get('bus', None)
        self.name = data.get('name', f'id:{self.id}')
        if self._has_poll_rate:
            self._pollRate = data['attributes']['pollRate']
        self.SetStatus(None, self.STATUS_RUNNING)

    def set_broker(self, mqttc, send_interface):
        ManagerOfBroker.__init__(self, mqttc)
        self.send_interface = send_interface

    def start(self):
        self.subscribe(mqtt.project.TopicProject.managers(self.project_id))

    def on_state(self, topic: mqtt.spread.topic.SpreadTopic = None, variable: mqtt.spread.variable.Variable = None):
        if topic and variable:
            if self.ready_timer:
                self.ready_timer.cancel()
            if isinstance(topic.entity_address, mqtt.spread.address.ProviderAddress):
                _p_id = topic.entity_address.id
                if _p_id in self._providers:
                    if variable.value is not None:
                        if topic.entity_address.funit_type not in self._providers[_p_id].not_repaired():
                            if not variable.preliminary:
                                self._providers[_p_id]._invalid_values[topic.entity_address.funit_type] = variable.value

        self.set_ready_timer()

    def set_ready_timer(self):
        if isinstance(self.ready_timer, threading.Timer):
            self.ready_timer.cancel()

        self.ready_timer = threading.Timer(1, self.on_ready)
        self.ready_timer.setName('on_ready Timer')
        self.ready_timer.start()

    def on_project(self, topic, data):
        self.unsubscribe(topic)

        if isinstance(data, (bytes, bytearray)):
            data = data.decode()

        payload = json.loads(data)
        if mqtt.project.TopicProject.MANAGERS in payload:
            for item in payload[mqtt.project.TopicProject.MANAGERS]:
                if item['id'] == self.id:
                    try:
                        self.parse(item)
                    except BaseException as ex:
                        raise InitError(str(ex))
                    self.subscribe(mqtt.project.TopicProject.providers(self.project_id), True)
                    return
            raise ProjectError("В файле 'managers.json' отсутствует менеджер с ID={}".format(self.id))
        elif mqtt.project.TopicProject.PROVIDERS in payload:
            for item in payload[mqtt.project.TopicProject.PROVIDERS]:
                if item['managerID'] == self.id:
                    try:
                        self.add_provider(item, self)
                    except BaseException as ex:
                        raise InitError(f'{ex} for {item}')
            logging.info('Project({}) loaded'.format(self.project_id))
            self.on_project_ready()

    def on_project_ready(self):
        self.on_state()
        self.subscribe(mqtt.spread.topic.State(self.get_address('+/+/#')))

    def add_provider(self, data, manager):
        provider = generator.generate_provider(data)
        provider.set_manager(manager)
        self._providers[provider.id] = provider

    def on_ready(self):
        self.unsubscribe(mqtt.spread.topic.State(self.get_address('+/+/#')))
        self.subscribe(mqtt.spread.topic.Set(self.get_address(None)))
        self.subscribe(mqtt.spread.topic.Renew(self.get_address(None)))

        if self.send_interface:
            self.send_interface.start()

        threading.Thread(target=self.info, name='InfoThread').start()

    def on_exit(self):
        super(Manager, self).on_exit()
        self.send_interface.set_stopped()

        self._stopped = True
        self.actions = {CONTROL: [], COMMAND: [], TUNER: [], INFO: [], SURVEY: []}

        if self.mqttc.connected:
            for p_id in self._providers:
                provider = self._providers[p_id]
                provider.on_exit()

            self.SetStatus(None, self.STATUS_STOPPED)

        if self._survey_timer:
            self._survey_timer.cancel()
        self._survey_event.set()
        self._step_event.set()
        self._event.set()
        self._external_cmd.set()

    def survey(self):
        while self._stopped is False:
            self._survey_event.wait()
            if len(self.actions[SURVEY]) == 0:
                for p_id in self._providers:
                    provider = self._providers[p_id]
                    if provider.surveyable:
                        provider.on_survey()
            if not self._step_event.is_set():
                self._step_event.set()
            self._survey_event.clear()
            if not self._stopped and self._pollRate:
                self._survey_timer = threading.Timer(interval=self._pollRate / 1000, function=self.dispatch_survey)
                self._survey_timer.setName('dispatch_survey Timer')
                self._survey_timer.start()

    def dispatch_survey(self):
        if not self._survey_event.is_set():
            self._survey_event.set()

    def external_command_handler(self):
        while self._stopped is False:
            ext = self.send_interface.get_ex_cmd()
            if ext:
                if not isinstance(ext, (tuple, list)):
                    ext = (ext, )
                try:
                    self.on_external_command(*ext)
                except BaseException as ex:
                    logging.debug(f'EXTERNAL FRAME ERROR[{ext}]')
                    logging.exception(ex)
            else:
                self._external_cmd.clear()
                self._external_cmd.wait()

    def on_external_command(self, *args):
        raise NotImplementedError()

    def event_handler(self):
        while not self._stopped:
            event = self.send_interface.get_event()
            if event:
                try:
                    self.on_event(event)
                except BaseException as ex:
                    logging.debug(f'EVENT FRAME ERROR[{event}]')
                    logging.exception(ex)

            else:
                self._event.clear()
                self._event.wait()

    def on_event(self, event):
        raise NotImplementedError()

    def info(self):
        self.SetStatus(None, self.STATUS_INFO)
        for p_id in self._sorted_info():
            if self._stopped:
                return
            provider = self._providers[p_id]
            if provider.infoble:
                try:
                    provider.get_bindings()
                    provider.check_valid()
                    provider.commit_states(provider.is_valid)
                    if provider.is_valid:
                        provider.get_info()
                        provider.override_info()
                except InvalidDevice as ex:
                    logging.warning(ex)
                    provider.set_valid(provider.INVALID)

        threading.Thread(target=self.survey, name='SurveyThread').start()
        threading.Thread(target=self.do_next, name='DoNextThread').start()
        threading.Thread(target=self.external_command_handler, name='ExternalThread').start()
        threading.Thread(target=self.event_handler, name='EventThread').start()
        self._survey_event.set()
        self.SetStatus(None, self.STATUS_SURVEY)

    def do_next(self):
        while self._stopped is False:
            self._step_event.wait()

            _priority = -1
            stack = ()
            for _priority, stack in self.actions.items():
                if stack: break

            while len(stack) > 0:
                need_break = False
                for _p, s in self.actions.items():
                    if _p > _priority and len(s) > 0:
                        need_break = True
                        break

                if need_break: break

                provider, cmd = stack.pop(0)

                try:
                    if isinstance(cmd, MultiCommand):
                        cmd.call()
                    elif provider:
                        provider.request(cmd)
                    else:
                        self.send_interface.send(cmd)
                except BaseException as ex:
                    logging.exception(ex)

                if _priority == SURVEY and self.INTERRUPT_SURVEY:
                    break

            self._step_event.clear()

    def add_command(self, provider, *cmds, is_control=False, is_tuner=False):
        if not self._stopped:
            for cmd in cmds:
                if cmd:
                    if is_control:
                        self.actions[CONTROL].append((provider, cmd))
                    elif is_tuner:
                        self.actions[TUNER].append((provider, cmd))
                    else:
                        self.actions[COMMAND].append((provider, cmd))
                else:
                    logging.warning('Empty command')

            if not self._step_event.is_set():
                self._step_event.set()

    def add_info(self, provider, cmd):
        if self._stopped:
            return
        if (provider, cmd) not in self.actions[INFO]:
            self.actions[INFO].append((provider, cmd))

    def add_survey(self, provider, cmd):
        if self._stopped:
            return
        if (provider, cmd) not in self.actions[SURVEY]:
            self.actions[SURVEY].append((provider, cmd))

    def on_command(self, topic: mqtt.spread.topic.Set, variable: mqtt.spread.variable.Variable):
        self.log_command(topic, variable)
        if self._stopped:
            logging.warning('Manager on scan/extend/construct. Reject command!')
            return
        try:
            if topic.entity_address.type == self.__class__.__name__ and topic.entity_address.id == self.id:
                super(Manager, self).on_command(topic, variable)
            elif topic.entity_address.id in self._providers:
                provider = self._providers[topic.entity_address.id]
                if provider.is_valid or (isinstance(topic, mqtt.spread.topic.Set) and topic.entity_address.funit_type == 'Binding'):
                    provider.on_command(topic, variable)
                else:
                    raise CommandGenerateError('device is invalid!')
            else:
                raise JocketError('объект с адресом {} не найден!'.format(topic.entity_address.id))
        except (JocketError, CommandGenerateError, ClassifierError, InvalidDevice, CommandError) as ex:
            logging.warning(ex)
            if variable.session_id:
                to_topic = mqtt.spread.topic.Reply(self.project_id, variable.session_id)
                self.publish(to_topic, str(ex))

        if not self._step_event.is_set():
            self._step_event.set()

    def log_command(self, topic: mqtt.spread.topic.SpreadTopic, variable: mqtt.spread.variable.Variable):
        logging.debug(
            f'COMMAND({variable.session_id if variable.session_id else ""}): ' +
            f'{topic.__class__.__name__} {topic.entity_address.type}({topic.entity_address.id})'
            f'.{topic.entity_address.funit_type}' +
            (f'={variable.value}' if isinstance(topic, mqtt.spread.topic.Set) else '') +
            (f' params: {variable.params}' if variable.params else '') +
            f' queue: {"+".join([str(len(q)) for q in self.actions.values()])}'
        )

    def SetStatus(self, value, sig=None):
        if self.send_interface:
            self.on_update('Status', value)

    def get_funit(self, funit_type):
        if self.__class__.__name__ in mqtt.classifier:
            if funit_type in mqtt.classifier[self.__class__.__name__]:
                return mqtt.classifier[self.__class__.__name__][funit_type]
        return None

    def get_address(self, funit_type):
        if funit_type is None: funit_type = '#'
        return mqtt.spread.address.ManagerAddress.of(self, funit_type)
