import logging
import threading

from spread_core import mqtt
from spread_core.errors.project_errors import InitError
from . import commands


class KnxInterface:
    def __init__(self, mqtt_client, timeout=1):
        self._mqtt_client = mqtt_client
        self._stopped = False
        self._ready = False
        self._send_topic = None
        self._dump_topic = None
        self.event_handler = None
        self.external_cmd_handler = None
        self._request_event = threading.Event()
        self.response = None
        self.request = None
        self.response_address = None
        self.timeout = timeout

    def set_bus_params(self, bus: str):
        self._send_topic = str(mqtt.bus.TopicBus.of_send(bus))
        self._dump_topic = str(mqtt.bus.TopicBus.of_dump(bus))

    def start(self):
        if self._dump_topic:
            logging.debug('Subscribed to %s' % self._dump_topic)
            self._mqtt_client.subscribe(str(self._dump_topic))
        else:
            raise InitError('SendTopic not set!')

    def set_event_handler(self, handler: iter):
        self.event_handler = handler

    def set_external_cmd_handler(self, handler: iter):
        self.external_cmd_handler = handler

    def send(self, command: commands.KnxFrame, state_address: str) -> commands.DPTU:
        logging.debug('-> %s [%s]' % (command, ' '.join([hex(b)[2:].rjust(2, '0').upper() for b in command.frame])))
        self.response = command.dptu.empty()
        self.response_address = state_address
        self.request = command
        self._mqtt_client.publish(self._send_topic, command.frame_str)

        self._request_event.wait(self.timeout)
        self._request_event.clear()
        yield self.response
        self.response = None

    def set_stopped(self):
        self._stopped = True

        logging.debug('Unsubscribed from %s' % self._dump_topic)
        self._mqtt_client.unsubscribe(str(self._dump_topic))

        if self.event_handler:
            try:
                self.event_handler.throw(StopIteration)
            except StopIteration:
                pass

        if self.external_cmd_handler:
            try:
                self.external_cmd_handler.throw(StopIteration)
            except StopIteration:
                pass

        if self._request_event and not self._request_event.is_set():
            self._request_event.set()

    def on_dump(self, response, *args, **kwargs):
        data, *flags = response.decode().split('#')
        frame = commands.KnxFrame.of(data)

        logging.debug('  <- %s [%s]' % (frame, ' '.join([hex(b)[2:].rjust(2, '0').upper() for b in bytes.fromhex(data)])))

        if self.response and frame.address == self.response_address:
            self.response.fill(frame.dptu.frame)
            self._request_event.set()
        else:
            self.event_handler.send(frame)
